/* usmb - mount SMB shares via FUSE and Samba
 * Copyright (C) 2006-2008 Geoff Johnstone
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/time.h>        // struct timeval needed by libsmbclient.h
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <libsmbclient.h>
#include <fuse.h>
#include <assert.h>
#include <errno.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "conffile.h"
#include "options.h"
#include "usmb.h"
#include "usmb_dir.h"
#include "usmb_file.h"
#include "utils.h"
#include "version.h"


SMBCCTX *ctx;
static char *server, *share, *mountpoint, *options,
            *domain, *username, *password;


char * make_url (const char *path)
{
  assert (NULL != share);

  if ((NULL == path) || ('\0' == path[0]))
    return xstrdup (share);
  else
    return concat_strings (2, share, path);
}


static inline void do_strncpy (char *to, const char *from, int tolen)
{
  strncpy (to, from, tolen);
  to[tolen - 1] = '\0';
}


static void auth_fn (const char *srv, const char *shr, char *wg, int wglen,
                     char *un, int unlen, char *pw, int pwlen)
{
  (void)srv;
  (void)shr;
  DEBUG (fprintf (stderr, "Authenticating for \\\\%s\\%s\n", srv, shr));
  DEBUG (fprintf (stderr, "Domain: %s; User: %s; Password:%s\n",
                  domain, username, password));

  if (NULL != domain)
    do_strncpy (wg, domain, wglen);

  do_strncpy (un, username, unlen);
  do_strncpy (pw, password, pwlen);
}


static bool create_smb_context (char *domain, char *username, SMBCCTX **pctx)
{
  *pctx = smbc_new_context();

  if (NULL == *pctx)
  {
    perror ("Cannot create SMB context");
    return false;
  }

  (*pctx)->workgroup = domain;
  (*pctx)->user = username;
  (*pctx)->timeout = 5000;
  (*pctx)->callbacks.auth_fn = auth_fn;

  if (NULL == smbc_init_context (*pctx))
  {
    perror ("Cannot initialise SMB context");
    smbc_free_context (*pctx, 1);
    return false;
  }

  return true;
}


#if 0
static int usmb_statfs (const char *path, struct statvfs *vfs)
{
  if ((NULL == path) || (NULL == vfs))
    return -EINVAL;

  memset (vfs, 0, sizeof (*vfs));

  vfs->f_bsize = 32768;
  vfs->f_blocks = 0x7FFFFFFFl;
  vfs->f_bfree = 0x7FFFFFFFl;
  vfs->f_bavail = 0x7FFFFFFFl;

  return 0;
}
#endif


static void * usmb_init (struct fuse_conn_info *conn)
{
  DEBUG (fputs ("usmb_init()\n", stderr));
  (void)conn;
  return NULL;
}


static void usmb_destroy (void *unused)
{
  (void)unused;
  DEBUG (fputs ("usmb_destroy()\n", stderr));
}


// probably won't (can't ?) implement these:
// readlink mknod symlink flush fsync

// fuse.h says "Just a placeholder, don't set".
// statfs

// no easy way of implementing these.
// access ftruncate

#ifdef __lint
#define SET_ELEMENT(name,value) value
#else
#define SET_ELEMENT(name,value) name = value
#endif
static struct fuse_operations fuse_ops = {
  SET_ELEMENT (.getattr, usmb_getattr),
  SET_ELEMENT (.readlink, NULL),
  SET_ELEMENT (.getdir, NULL),
  SET_ELEMENT (.mknod, NULL),
  SET_ELEMENT (.mkdir, usmb_mkdir),
  SET_ELEMENT (.unlink, usmb_unlink),
  SET_ELEMENT (.rmdir, usmb_rmdir),
  SET_ELEMENT (.symlink, NULL),
  SET_ELEMENT (.rename, usmb_rename),
  SET_ELEMENT (.link, NULL),
  SET_ELEMENT (.chmod, usmb_chmod),
  SET_ELEMENT (.chown, NULL), // usmb_chown, --not implemented in libsmbclient
  SET_ELEMENT (.truncate, usmb_truncate),
  SET_ELEMENT (.utime, usmb_utime),
  SET_ELEMENT (.open, usmb_open),
  SET_ELEMENT (.read, usmb_read),
  SET_ELEMENT (.write, usmb_write),
  SET_ELEMENT (.statfs, NULL),
  SET_ELEMENT (.flush, NULL),
  SET_ELEMENT (.release, usmb_release),
  SET_ELEMENT (.fsync, NULL),
  SET_ELEMENT (.setxattr, usmb_setxattr),
  SET_ELEMENT (.getxattr, usmb_getxattr),
  SET_ELEMENT (.listxattr, usmb_listxattr),
  SET_ELEMENT (.removexattr, usmb_removexattr),
  SET_ELEMENT (.opendir, usmb_opendir),
  SET_ELEMENT (.readdir, usmb_readdir),
  SET_ELEMENT (.releasedir, usmb_releasedir),
  SET_ELEMENT (.fsyncdir, NULL),
  SET_ELEMENT (.init, usmb_init),
  SET_ELEMENT (.destroy, usmb_destroy),
  SET_ELEMENT (.access, NULL),
  SET_ELEMENT (.create, usmb_create),
  SET_ELEMENT (.ftruncate, NULL),
  SET_ELEMENT (.fgetattr, usmb_fgetattr),
  SET_ELEMENT (.lock, NULL),                   // TODO: implement
  SET_ELEMENT (.utimens, NULL),                // TODO: implement
  SET_ELEMENT (.bmap, NULL),                   // TODO: implement
};


// this should really open() the file and check the fd, but the XML parser
// takes a filename, not a file descriptor
static bool check_conf_perms (const char *conffile)
{
  struct stat buf;
  if (0 == stat (conffile, &buf))
  {
    if (getuid() != buf.st_uid)
    {
      fprintf (stderr, "You do not own the configuration file %s\n",
               conffile);
      return false;
    }

    if (buf.st_mode & (S_IRWXG | S_IRWXO))
    {
      fprintf (stderr, "Configuration file %s is accessible to non-owner\n",
               conffile);
      return false;
    }
  }
  else
  {
    fprintf (stderr, "Cannot stat configuration file %s: %s\n",
             conffile, strerror (errno));
    return false;
  }

  return true;
}


static bool create_share_name (const char *server, const char *sharename)
{
  size_t len = strlen ("smb:///") +
               strlen (server) +
               strlen (sharename) + 1;

  if (NULL == (share = malloc (len)))
  {
    perror ("Cannot allocate share name");
    return false;
  }

  snprintf (share, len, "smb://%s/%s", server, sharename);
  DEBUG (fprintf (stderr, "Share URL: %s\n", share));
  return true;
}


static void free_strings (char *sharename)
{
  xfree (sharename);
  xfree (password);
  xfree (username);
  xfree (domain);
  xfree (options);
  xfree (mountpoint);
  xfree (share);
  xfree (server);
}


int main (int argc, char **argv)
{
  const char *conffile, *mountid;
  char *sharename = NULL;

  {
    static char conf[256];
    snprintf (conf, sizeof (conf), "%s/.usmb.conf", getenv ("HOME"));
    conffile = conf;
  }

  if (!parse_args (&argc, &argv, &mountid, &conffile))
    return EXIT_FAILURE;

  if (!check_conf_perms (conffile) ||
      !conffile_get_mount (conffile, mountid,
                           &server, &sharename, &mountpoint, &options,
                           &domain, &username, &password))
    return EXIT_FAILURE;

  if (!create_share_name (server, sharename) ||
      !create_smb_context (domain, username, &ctx))
  {
    free_strings (sharename);
    return EXIT_FAILURE;
  }

  DEBUG (fprintf (stderr, "Username: %s\\%s\n", domain, username));
  show_about (stdout);

  int fuse_argc;
  char **fuse_argv;
  build_fuse_args (options, mountpoint, &fuse_argc, &fuse_argv);
  int ret = fuse_main (fuse_argc, fuse_argv, &fuse_ops, NULL);

  smbc_free_context (ctx, 1);
  free_strings (sharename);

  return ret;
}

