/* usmb - mount SMB shares via FUSE and Samba
 * Copyright (C) 2006-2008 Geoff Johnstone
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/time.h>        // struct timeval needed by libsmbclient.h
#include <libsmbclient.h>
#include <fuse.h>
#include <dirent.h>
#include <errno.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "usmb_dir.h"
#include "usmb.h"
#include "utils.h"


int usmb_mkdir (const char *dirname, mode_t mode)
{
  char *url = make_url (dirname);
  if (NULL == url)
    return -ENOMEM;

  DEBUG (fprintf (stderr, "mkdir (%s)\n", url));
  int ret = ctx->mkdir (ctx, url, mode) ? -errno : 0;
  free (url);
  return ret;
}


int usmb_rmdir (const char *dirname)
{
  char *url = make_url (dirname);
  if (NULL == url)
    return -ENOMEM;

  DEBUG (fprintf (stderr, "rmdir (%s)\n", url));
  int ret = ctx->rmdir (ctx, url) ? -errno : 0;
  free (url);
  return ret;
}


int usmb_opendir (const char *dirname, struct fuse_file_info *fi)
{
  char *url = make_url (dirname);
  if (NULL == url)
    return -ENOMEM;

  DEBUG (fprintf (stderr, "opendir (%s)", url));
  SMBCFILE *file = ctx->opendir (ctx, url);
  DEBUG (fprintf (stderr, " = %p\n", (void *)file));

  int ret = (NULL == file) ? -errno : 0;
  free (url);
  fi->fh = smbcfile_to_fd (file);

  return ret;
}


int usmb_readdir (const char *path, void *h, fuse_fill_dir_t filler,
                  off_t offset, struct fuse_file_info *fi)
{
  (void)path;
  (void)offset;

  struct smbc_dirent *dirent;
  errno = 0;
  SMBCFILE *file = fd_to_smbcfile (fi->fh);

  DEBUG (fprintf (stderr, "readdir (%s, %p)\n", path, (void *)file));

  while (NULL != (dirent = ctx->readdir (ctx, file)))
  {
    struct stat stbuf;

    switch (dirent->smbc_type)
    {
      case SMBC_DIR:
        stbuf.st_mode = DT_DIR << 12;
        break;

      case SMBC_FILE:
        stbuf.st_mode = DT_REG << 12;
        break;

      case SMBC_LINK:
        stbuf.st_mode = DT_LNK << 12;
        break;

      default:
        break;
    }

    if (1 == filler (h, dirent->name, &stbuf, 0))  /* if error */
      return -1;
  }

  return -errno;
}


int usmb_releasedir (const char *path, struct fuse_file_info *fi)
{
  (void)path;

  SMBCFILE *file = fd_to_smbcfile (fi->fh);
  DEBUG (fprintf (stderr, "releasedir (%s, %p)\n", path, (void *)file));
  return (ctx->closedir (ctx, file) < 0) ? -errno : 0;
}


int usmb_setxattr (const char *path, const char *name, const char *value,
                   size_t size, int flags)
{
  char *url = make_url (path);
  if (NULL == url)
    return -ENOMEM;

  DEBUG (fprintf (stderr, "setxattr (%s, %s, %p, %u, %x)\n",
                  path, url, value, size, flags));
  int ret = ctx->setxattr (ctx, url, name, value, size, flags) ? -errno : 0;
  free (url);
  return ret;
}


int usmb_getxattr (const char *path, const char *name, char *value, size_t size)
{
  char *url = make_url (path);
  if (NULL == url)
    return -ENOMEM;

  DEBUG (fprintf (stderr, "getxattr (%s, %s, %p, %u)\n",
                  path, url, value, size));
  int ret = ctx->getxattr (ctx, url, name, value, size) ? -errno : 0;
  free (url);
  return ret;
}


int usmb_listxattr (const char *path, char *list, size_t size)
{
  char *url = make_url (path);
  if (NULL == url)
    return -ENOMEM;

  DEBUG (fprintf (stderr, "listxattr (%s, %p, %u)\n", url, list, size));
  int ret = ctx->listxattr (ctx, url, list, size) ? -errno : 0;
  free (url);
  return ret;
}


int usmb_removexattr (const char *path, const char *name)
{
  char *url = make_url (path);
  if (NULL == url)
    return -ENOMEM;

  DEBUG (fprintf (stderr, "removexattr (%s, %s)\n", url, name));
  int ret = ctx->removexattr (ctx, url, name) ? -errno : 0;
  free (url);
  return ret;
}

