/* usmb - mount SMB shares via FUSE and Samba
 * Copyright (C) 2006-2008 Geoff Johnstone
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/time.h>        // struct timeval needed by libsmbclient.h
#include <libsmbclient.h>
#include <assert.h>
#include <fuse.h>
#include <errno.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "usmb_file.h"
#include "usmb.h"
#include "utils.h"


int usmb_getattr (const char *filename, struct stat *st)
{
  char *url = make_url (filename);
  if (NULL == url)
    return -ENOMEM;

  DEBUG (fprintf (stderr, "stat (%s)\n", url));

  int ret = ctx->stat (ctx, url, st);
  free (url);
  return (ret < 0) ? -errno : 0;
}


int usmb_fgetattr (const char *filename, struct stat *st,
                   struct fuse_file_info *fi)
{
  (void)filename;
  SMBCFILE *file = fd_to_smbcfile (fi->fh);
  DEBUG (fprintf (stderr, "fgetattr (%s, %p)\n", filename, (void *)file));
  return (ctx->fstat (ctx, file, st) < 0) ? -errno : 0;
}


int usmb_unlink (const char *filename)
{
  char *url = make_url (filename);
  if (NULL == url)
    return -ENOMEM;

  DEBUG (fprintf (stderr, "unlink (%s)\n", url));
  int ret = (ctx->unlink (ctx, url) < 0) ? -errno : 0;
  free (url);
  return ret;
}


int usmb_open (const char *filename, struct fuse_file_info *fi)
{
  char *url = make_url (filename);
  if (NULL == url)
    return -ENOMEM;

  DEBUG (fprintf (stderr, "open (%s, %d)", url, fi->flags));
  SMBCFILE *file = ctx->open (ctx, url, fi->flags, 0);
  DEBUG (fprintf (stderr, " = %p\n", (void *)file));

  free (url);
  int ret = (NULL == file) ? -errno : 0;
  fi->fh = smbcfile_to_fd (file);

  return ret;
}


int usmb_release (const char *filename, struct fuse_file_info *fi)
{
  (void)filename;
  SMBCFILE *file = fd_to_smbcfile (fi->fh);
  DEBUG (fprintf (stderr, "release (%s, %p)\n", filename, (void *)file));
  return (ctx->close_fn (ctx, file) < 0) ? -errno : 0;
}


int usmb_read (const char *filename, char *buff, size_t len, off_t off,
               struct fuse_file_info *fi)
{
  (void)filename;
  (void)off;

  assert (len <= 32768);

  SMBCFILE *file = fd_to_smbcfile (fi->fh);
  DEBUG (fprintf (stderr, "read (%p, %p, %u, %lld) ",
                  (void *)file, buff, len, off));

  if (ctx->lseek (ctx, file, off, SEEK_SET) < 0)
  {
    fprintf (stderr, "- seek failed: %d\n", -errno);
    return -errno;
  }

  int bytes = ctx->read (ctx, file, buff, len);
  DEBUG (fprintf (stderr, "= %d\n", bytes));
  return (bytes < 0) ? -errno : (int)bytes;
}


int usmb_write (const char *filename, const char *buff, size_t len, off_t off,
                struct fuse_file_info *fi)
{
  (void)filename;
  (void)off;

  SMBCFILE *file = fd_to_smbcfile (fi->fh);
  DEBUG (fprintf (stderr, "write (%p, %p, len=%u, off=%lld) ",
                  (void *)file, buff, len, off));

  if (ctx->lseek (ctx, file, off, SEEK_SET) < 0)
  {
    fprintf (stderr, "- seek failed: %d\n", -errno);
    return -errno;
  }

  size_t written = 0;
  int bytes = 0;

  // No idea whether Windows servers don't like > 32768 byte writes
  // (cf. usmb_read), but taking no chances...
  while (written < len)
  {
    bytes = ctx->write (ctx, file, (char *)buff, (len > 32768) ? 32768 : len);

    if (bytes < 0)
      break;

    written += bytes;
    buff += bytes;

    // avoids infinite loops
    if (0 == bytes)
      break;
  }

  DEBUG (fprintf (stderr, "= %d\n", (bytes < 0) ? -errno : (int)written));
  return (bytes < 0) ? -errno : (int)written;
}


int usmb_create (const char *filename, mode_t mode, struct fuse_file_info *fi)
{
  char *url = make_url (filename);
  if (NULL == url)
    return -ENOMEM;

  DEBUG (fprintf (stderr, "creat (%s)", url));

  SMBCFILE *file = ctx->creat (ctx, url, mode);
  DEBUG (fprintf (stderr, " = %p\n", (void *)file));
  int ret = (NULL == file) ? -errno : 0;
  fi->fh = smbcfile_to_fd (file);
  
  free (url);
  return ret;
}


int usmb_rename (const char *from, const char *to)
{
  char *fromurl = make_url (from);
  if (NULL == fromurl)
    return -ENOMEM;

  char *tourl = make_url (to);
  if (NULL == tourl)
  {
    free (fromurl);
    return -ENOMEM;
  }

  DEBUG (fprintf (stderr, "rename (%s, %s)\n", fromurl, tourl));
  int ret = (ctx->rename (ctx, fromurl, ctx, tourl) < 0) ? -errno : 0;
  free (tourl);
  free (fromurl);
  return ret;
}


int usmb_utime (const char *filename, struct utimbuf *utb)
{
  struct utimbuf tmp_utb;

  if (NULL == utb)
  {
    for (;;)
    {
      time_t now = time (NULL);
      if ((time_t)-1 != now)
      {
        tmp_utb.actime = tmp_utb.modtime = now;
        break;
      }

      if (EINTR != errno)
        return -errno;
    }

    utb = &tmp_utb;
  }

  char *url = make_url (filename);
  if (NULL == url)
    return -ENOMEM;

  struct timeval tv[2] = {
    { .tv_sec = utb->actime,  .tv_usec = 0 },
    { .tv_sec = utb->modtime, .tv_usec = 0 },
  };

  DEBUG (fprintf (stderr, "utime (%s)\n", url));
  int ret = (ctx->utimes (ctx, url, tv) < 0) ? -errno : 0;
  free (url);
  return ret;
}


#if 0
Samba defines utimes as taking struct timevals rather than timespecs.
int usmb_utimes (const char *filename, const struct timespec ts[2])
{
  char *url = make_url (filename);
  if (NULL == url)
    return -ENOMEM;

  DEBUG (fprintf (stderr, "utimes (%s)\n", url));
  int ret = (ctx->utimes (ctx, url, ts) < 0) ? -errno : 0;
  free (url);
  return ret;
}
#endif


int usmb_truncate (const char *filename, off_t offset)
{
  /* Windows doesn't support truncation so we implement a limited version:
   *  0 == offset => create a new file for writing.
   *  current size == offset => succeed.
   *  else return -ENOSYS.
   */

  DEBUG (fprintf (stderr, "truncate (%s, %llu)\n", filename, offset));

  if (NULL == filename)
    return -EINVAL;

  if (0 == offset)
  {
    char *url = make_url (filename);
    if (NULL == url)
      return -ENOMEM;
  
    SMBCFILE *file = ctx->open (ctx, url, O_WRONLY | O_TRUNC, 0);
    if (NULL == file)
    {
      free (url);
      return -errno;
    }
  
    ctx->close_fn (ctx, file);
    free (url);

    return 0;
  }

  struct stat st;
  int ret = usmb_getattr (filename, &st);

  if (0 != ret)
    return ret;

  return (offset == st.st_size) ? 0 : -ENOSYS;
}


int usmb_chmod (const char *filename, mode_t mode)
{
  char *url = make_url (filename);
  if (NULL == url)
    return -ENOMEM;
  
  DEBUG (fprintf (stderr, "chmod (%s, %u)\n", url, mode));
  int ret = (ctx->chmod (ctx, url, mode) < 0) ? -errno : 0;
  free (url);
  return ret;
}


#if 0
// smbc_chown isn't implemented in libsmbclient
int usmb_chown (const char *filename, uid_t owner, uid_t group)
{
  char *url = make_url (filename);
  if (NULL == url)
    return -ENOMEM;
  
  DEBUG (fprintf (stderr, "chown (%s, %d, %d)\n", url, owner, group));
  int ret = (ctx->chown (ctx, url, owner, group) < 0) ? -errno : 0;
  free (url);
  return ret;
}
#endif
